# =======================================================
# Functions for retrieving data from the database
# =======================================================

import sqlite3
import tabulate
import time
import datetime


# Returns a ready to print table, based on provided
# date range and mode
def report(date_range, mode):
    conn = sqlite3.connect("activities.db")
    cursor = conn.cursor()

    table_headers = {
        "total": ["Name", "Total screen time"],
        "avg": ["Name", "Average screen time"],
        "raw": ["Id", "Date", "Process name", "Start", "End"]
    }

    # Ensure the table is created
    cursor.execute("CREATE TABLE IF NOT EXISTS activities ("
                   " id INTEGER PRIMARY KEY,"
                   " date TEXT NOT NULL,"
                   " process_name TEXT NOT NULL,"
                   " start INTEGER NOT NULL,"
                   " end INTEGER NOT NULL)")

    if mode == "total":
        # Total time in seconds per process name in range
        # If we just sum up all screen time in range (end - start), SQLite will cap it at 24 hours
        # because it treats the screen time as datetime.
        total = list(cursor.execute("SELECT process_name, "
                                    "SUM(STRFTIME('%s', end) - STRFTIME('%s', start))"
                                    "FROM activities "
                                    "WHERE date BETWEEN ? AND ?  "
                                    "GROUP BY process_name "
                                    "ORDER BY "
                                    "SUM(STRFTIME('%s', end) - STRFTIME('%s', start)) DESC",
                                    (date_range[0], date_range[1])))

        result = []
        for row in total:
            result.append((row[0], datetime.timedelta(seconds=int(row[1]))))

    elif mode == "avg":
        # Get total time in seconds per process name in range, then divide them
        # by a number of days during which the user had activity.
        # It's easier to postprocess the results, than doing this in pure SQL.
        total = list(cursor.execute("SELECT process_name, "
                                    "SUM(STRFTIME('%s', end) - STRFTIME('%s', start)) "
                                    "FROM activities "
                                    "WHERE date BETWEEN ? AND ?"
                                    "GROUP BY process_name "
                                    "ORDER BY "
                                    "SUM(STRFTIME('%s', end) - STRFTIME('%s', start)) DESC",
                                    (date_range[0], date_range[1])))

        # Get number of days to average on
        days = int(list(cursor.execute("SELECT COUNT(DISTINCT date) FROM activities WHERE date BETWEEN ? AND ?",
                                       (date_range[0], date_range[1])))[0][0])

        # Calculate the average time per process name and convert to HH:MM:SS
        result = []
        for row in total:
            result.append((row[0], datetime.timedelta(seconds=int(row[1]) // days)))

    elif mode == "raw":
        # Pull out the raw records, nothing fancy
        result = cursor.execute("SELECT * "
                                "FROM activities "
                                "WHERE date BETWEEN ? AND ?",
                                (date_range[0], date_range[1]))

        # Convert an iterator to a list
        result = list(result)


    else:
        raise Exception("Invalid report mode:", mode)

    # If there is data
    if result:
        # Format the data with tabulate
        table = tabulate.tabulate(result, tablefmt="fancy_grid", headers=table_headers[mode])
        return table
    else:
        return None



# =======================================================
# Provides the track() function, that tracks time spent
# on apps and saves it to activities.db
# =======================================================
import tabulate
import win32gui
import psutil
import win32process
import sqlite3
import datetime
import time
import signal

from .config import *


interrupted = False


def track(debug=False):

    # Returns the process name of the foreground window
    def get_active_window_process_name():
        pid = win32process.GetWindowThreadProcessId(win32gui.GetForegroundWindow())
        try:
            process_name = psutil.Process(pid[-1]).name()
            return process_name

        except (psutil.NoSuchProcess, ValueError, ProcessLookupError):
            return None

    # We'll count the activities to print them later. More feedback -> better UX
    activity_count = 0

    # Connect the database
    conn = sqlite3.connect("activities.db")
    cursor = conn.cursor()

    # A signal handler. When the user hits Ctrl+C, the script will only
    # exit after the data is saved
    def signal_handler(signal, frame):
        global interrupted
        interrupted = True
        print("Stopping...")

    signal.signal(signal.SIGINT, signal_handler)

    # Ensure the table is created
    cursor.execute("CREATE TABLE IF NOT EXISTS activities ("
                   " id INTEGER PRIMARY KEY,"
                   " date TEXT NOT NULL,"
                   " process_name TEXT NOT NULL,"
                   " start INTEGER NOT NULL,"
                   " end INTEGER NOT NULL)")

    print("Recording...\nHit Ctrl+C to stop.")

    # This is to get the current process when the script has started
    process_name_1 = get_active_window_process_name()
    while not process_name_1 or process_name_1 in IGNORE_PROCESSES:
        process_name_1 = get_active_window_process_name()

    if debug:
        print(process_name_1)

    date = datetime.date.today()

    start = datetime.datetime.now().strftime("%H:%M:%S")

    # Poll for the current process name and detect when it changes
    while True:

        # If date changed, update it and split the activity (start -> 23:59:59 and 00:00:00 -> end)
        new_date = datetime.date.today()
        if date < new_date:

            cursor.execute("INSERT INTO activities (date, process_name, start, end) VALUES (?, ?, ?, ?)",
                           (date, process_name_1, start, "23:59:59"))

            date = new_date

            conn.commit()

            if debug:
                print(f"Date changed. Carrying over {process_name_1} activity.")
                print(f"{process_name_1} {start} -> {process_name_1} 23:59:59")

            start = "00:00:01"

        process_name_2 = get_active_window_process_name()
        while not process_name_2:
            process_name_2 = get_active_window_process_name()

        # If process changed, add the activity into database
        if process_name_1 != process_name_2 and process_name_2 not in IGNORE_PROCESSES:

            end = datetime.datetime.now().strftime("%H:%M:%S")

            if debug:
                print(f"{process_name_1} {start} -> {process_name_2} {end}")

            # Update database, commit changes
            cursor.execute("INSERT INTO activities (date, process_name, start, end) VALUES (?, ?, ?, ?)",
                           (date, process_name_1, start, end))
            conn.commit()

            process_name_1 = process_name_2
            start = end
            activity_count += 1

        # If the user hits Ctrl+C, save changes and abort the loop
        if interrupted:
            cursor.execute("INSERT INTO activities (date, process_name, start, end) VALUES (?, ?, ?, ?)",
                           (date, process_name_1, start, datetime.datetime.now().strftime("%H:%M:%S")))
            conn.commit()
            activity_count += 1
            print(f"Recording stopped, {activity_count} activities added into the database.")
            return 0

        time.sleep(POLLING_DELAY)


# If track.py is ran as a script
if __name__ == "__main__":
    track(debug=True)



# =======================================================
# CLI interface for the app
# =======================================================
from datetime import date
import click
import sqlite3
import re
import sys

from .track import track
from .report import report


@click.group()
def cli():
    pass


@cli.command(name="track", help="Track your activities and save them into a database. ")
@click.option("--debug", is_flag=True, help="Print the transitions between processes (1.exe 11:00 -> 2.exe 12:00)")
def track_(debug):
    track(debug)


@cli.command(name="report", help="Create a report of your activities.")
@click.argument("date_range", default=f"{date.today()}:{date.today()}")
@click.option("--mode", "-m", default="total", type=click.Choice(["total", "avg", "raw"]),
              help="One of the three modes: total, avg and raw. Default is total.")
def report_(date_range, mode):  # _ to avoid name collision
    date_regex = "^\d{4}-(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01])$"
    date_range_regex = "^\d{4}-(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01])\:\d{4}-(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01])$"

    # Validate the provided date
    if re.search(date_regex, date_range):
        # If given a single date, convert to a range
        date_range = [date_range, date_range]
    elif re.search(date_range_regex, date_range):
        # If given a range, split into a list
        date_range = date_range.split(":")
    else:
        print(f"Error: Invalid date (range) '{date_range}'."
              f" Please use one of the following formats: 'YYYY-MM-DD' or 'YYYY-MM-DD:YYYY-MM-DD'")
        sys.exit()

    # Print results
    print(f"Results for {':'.join(date_range) if date_range[0] != date_range[1] else date_range[0]} in {mode} mode:")
    result = report(date_range, mode)
    print(result if result else "No data. Looks like there are no activities in this range!")


@cli.command(name="clear_db", help="Clear the activity database.")
def clear_db():
    if click.confirm("Do you want to clear the database? You won't be able to restore it."):
        conn = sqlite3.connect("activities.db")
        cursor = conn.cursor()
        rows_count = int(list(cursor.execute("SELECT COUNT(id) FROM activities"))[0][0])
        cursor.execute("DROP TABLE activities")
        print(rows_count, "row(s) deleted succesfully.")
        conn.close()


if __name__ == '__main__':
    cli()



# How often the script will check if the process changed, in seconds
POLLING_DELAY = 0.2

# These processes randomly pop up when idling, the script will ignore
# them. You can add your own of similar nature (a reminder or an alarm,
# a floating dock, any process that's sometimes or always in the foreground,
# but you don't want to track its screen time)
IGNORE_PROCESSES = ["ShellExperienceHost.exe", "StartMenuExperienceHost.exe", "mbamtray.exe", "SearchApp.exe", "ueli.exe", "explorer.exe", "dwm.exe"]



from setuptools import setup, find_packages

setup(
    name='time-on-fire',
    version='0.1.1',
    author="zapvolt",
    author_email="ag11012008@gmail.com",
    description="A simple activity tracker for Windows.",
    license="GNU GPLv3",
    long_description_content_type="text/markdown",
    long_description=open("README.md").read(),
    packages=["time_on_fire"],
    package_data={
        "time_on_fire": ["activities.db"],
    },
    include_package_data=True,
    install_requires=[
        "click", "pywin32", "psutil", "tabulate"
    ],
    keywords=["python time-tracking time tracking monitoring report"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Intended Audience :: End Users/Desktop",
        "Programming Language :: Python :: 3",
        "Operating System :: Microsoft :: Windows",
    ],
    entry_points={
        "console_scripts": [
            "tof=time_on_fire.cli:cli",
        ],
    },
)
